#!/usr/bin/env -S uv run --script
# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "openai",
#     "opentelemetry-api>=1.31.0",
#     "opentelemetry-sdk>=1.31.0",
#     "opentelemetry-exporter-otlp>=1.31.0",
#     "openinference-instrumentation-openai>=0.1.28,<0.2.0",
#     "httpx>=0.27.0"
# ]
# ///
import json
import os
import httpx
from openai import OpenAI

# OpenTelemetry imports
from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import Resource
from opentelemetry.semconv.resource import ResourceAttributes

# Import OpenInference instrumenter for OpenAI
from openinference.instrumentation.openai import OpenAIInstrumentor

# Configure OpenTelemetry
resource = Resource(
    attributes={
        ResourceAttributes.SERVICE_NAME: "openai-weather-app",
        ResourceAttributes.SERVICE_VERSION: "0.1.0",
    }
)

# Initialize the trace provider with the resource
trace_provider = TracerProvider(resource=resource)

# If OTEL_EXPORTER_OTLP_ENDPOINT is not set, we'll use console exporter
# Otherwise, we'll use OTLP exporter for sending to the Patronus collector
if os.environ.get("OTEL_EXPORTER_OTLP_ENDPOINT"):
    # Configure OTLPSpanExporter
    # The environment variables OTEL_EXPORTER_OTLP_ENDPOINT and OTEL_EXPORTER_OTLP_HEADERS
    # should be set before running this example
    otlp_exporter = OTLPSpanExporter()
    trace_provider.add_span_processor(BatchSpanProcessor(otlp_exporter))
else:
    # For local development/testing we can use ConsoleSpanExporter
    from opentelemetry.sdk.trace.export import ConsoleSpanExporter

    trace_provider.add_span_processor(BatchSpanProcessor(ConsoleSpanExporter()))

# Set the provider
trace.set_tracer_provider(trace_provider)

# Initialize OpenInference instrumenter for OpenAI
# This will automatically instrument all OpenAI API calls
openai_instrumentor = OpenAIInstrumentor()
openai_instrumentor.instrument()

# Get a tracer for our manual spans
tracer = trace.get_tracer("openai.weather.example")


def get_weather(latitude, longitude):
    """Get weather data from the Open Meteo API using httpx"""
    with tracer.start_as_current_span(
        "get_weather",
        attributes={
            "service.name": "weather_api",
            "weather.latitude": latitude,
            "weather.longitude": longitude,
        },
    ) as span:
        try:
            # Create the URL with parameters
            url = "https://api.open-meteo.com/v1/forecast"
            params = {
                "latitude": latitude,
                "longitude": longitude,
                "current": "temperature_2m,wind_speed_10m",
                "hourly": "temperature_2m,relative_humidity_2m,wind_speed_10m",
            }

            # Trace the HTTP request using httpx
            with tracer.start_as_current_span(
                "http_request",
                attributes={
                    "http.method": "GET",
                    "http.url": url,
                    "http.request.query": str(params),
                },
            ):
                # Use httpx client for the request
                with httpx.Client() as client:
                    response = client.get(url, params=params)

                # Add response information to the span
                span.set_attribute("http.status_code", response.status_code)

                if response.status_code != 200:
                    span.record_exception(
                        Exception(f"Weather API returned status {response.status_code}")
                    )
                    span.set_status(trace.StatusCode.ERROR)
                    return None

                data = response.json()
                temperature = data["current"]["temperature_2m"]

                # Add weather data to the span
                span.set_attribute("weather.temperature_celsius", temperature)

                return temperature
        except Exception as e:
            # Record the exception in the span
            span.record_exception(e)
            span.set_status(trace.StatusCode.ERROR, str(e))
            raise


def get_client():
    """Create and return an OpenAI client"""
    with tracer.start_as_current_span("get_openai_client"):
        return OpenAI()


def call_llm(client, user_prompt):
    """Call the OpenAI API to process the user prompt

    Note: With OpenInference instrumenter, the OpenAI API call will be
    automatically traced. This function adds some additional manual spans
    for demonstration purposes.
    """
    with tracer.start_as_current_span(
        "call_llm",
        attributes={
            "ai.prompt.text": user_prompt,
            "ai.prompt.tokens": len(user_prompt.split()),
        },
    ) as span:
        try:
            # Define tools available to the model
            tools = [
                {
                    "type": "function",
                    "name": "get_weather",
                    "description": "Get current temperature for provided coordinates in celsius.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "latitude": {"type": "number"},
                            "longitude": {"type": "number"},
                        },
                        "required": ["latitude", "longitude"],
                        "additionalProperties": False,
                    },
                    "strict": True,
                }
            ]

            input_messages = [{"role": "user", "content": user_prompt}]

            # The OpenAI API call will be automatically traced by OpenInference
            # We don't need to create a span for it, but we can add attributes to our parent span
            response = client.responses.create(
                model="gpt-4.1",
                input=input_messages,
                tools=tools,
            )

            # Check if the response contains a tool call
            has_tool_call = False
            if response.output and len(response.output) > 0:
                output = response.output[0]
                if output.type == "function_call":
                    has_tool_call = True
                    span.set_attribute("openai.response.tool_called", output.name)

            span.set_attribute("openai.response.has_tool_call", has_tool_call)

            return response
        except Exception as e:
            span.record_exception(e)
            span.set_status(trace.StatusCode.ERROR, str(e))
            raise


def main():
    """Main function to process the weather query"""
    with tracer.start_as_current_span("openai-weather-main") as root_span:
        user_prompt = "What's the weather like in Paris today?"
        root_span.set_attribute("query", user_prompt)

        try:
            client = get_client()
            response = call_llm(client, user_prompt)
            print("LLM Response")
            print(response.model_dump_json())

            weather_response = None
            if response.output:
                output = response.output[0]
                if output.type == "function_call" and output.name == "get_weather":
                    # Parse the arguments from the function call
                    with tracer.start_as_current_span(
                        "parse_function_call", attributes={"function_name": output.name}
                    ):
                        kwargs = json.loads(output.arguments)
                        root_span.set_attribute(
                            "weather.latitude", kwargs.get("latitude")
                        )
                        root_span.set_attribute(
                            "weather.longitude", kwargs.get("longitude")
                        )

                    print("Weather API Response")
                    weather_response = get_weather(**kwargs)
                    print(weather_response)

            if weather_response:
                with tracer.start_as_current_span("format_weather_response"):
                    print(user_prompt)
                    formatted_answer = f"Answer: {weather_response}"
                    print(formatted_answer)
                    root_span.set_attribute("weather.answer", formatted_answer)

            # Mark the trace as successful
            root_span.set_status(trace.StatusCode.OK)

        except Exception as e:
            # Record any exceptions that occurred
            root_span.record_exception(e)
            root_span.set_status(trace.StatusCode.ERROR, str(e))
            print(f"Error: {e}")


if __name__ == "__main__":
    main()
    # Ensure all spans are exported before the program exits
    trace_provider.shutdown()



from patronus.evals import evaluator, RemoteEvaluator
from patronus.experiments import run_experiment, Row, TaskResult, FuncEvaluatorAdapter


def my_task(row: Row, **kwargs):
    return f"{row.task_input} World"


# Reference remote Judge Patronus Evaluator with is-concise criteria.
# This evaluator runs remotely on Patronus infrastructure.
is_concise = RemoteEvaluator("judge", "patronus:is-concise")


@evaluator()
def exact_match(row: Row, task_result: TaskResult, **kwargs):
    print(f"{task_result.output=}  :: {row.task_output=}")
    return task_result.output == row.task_output


result = run_experiment(
    project_name="Tutorial Project",
    dataset=[
        {
            "task_input": "Hello",
            "gold_answer": "Hello World",
        },
    ],
    task=my_task,
    evaluators=[is_concise, FuncEvaluatorAdapter(exact_match)],
)

result.to_csv("./experiment.csv")



import abc
import asyncio
import functools
import typing
from typing import Optional


import pydantic
from opentelemetry import trace
from opentelemetry import _logs

# Would be imported from patronus.trace?
tracer = trace.get_tracer("patronus")
logger = _logs.get_logger("patronus")


class Evaluation(pydantic.BaseModel):
    pass_: Optional[bool] = None
    score: Optional[float] = None
    text_output: Optional[str] = None
    metadata: Optional[dict] = None
    tags: Optional[dict] = None

    # explanation: Optional[str] = None
    # evaluation_duration: Optional[datetime.timedelta] = None
    # explanation_duration: Optional[datetime.timedelta] = None


"""
wraps reports the function call as:
- evaluation *span* if it returns Evaluation object
- function *span* otherwise
"""


@patronus.wrap
def my_scorer(foo, bar, *, tar): ...


def _trace_evaluation(func):
    @functools.wraps(func)
    async def async_wrapper(*args, **kwargs):
        with tracer.start_as_current_span() as span:
            span: trace.Span
            span.set_attributes(
                {
                    "function_name": func.__name__,
                }
            )
            return await func(*args, **kwargs)
        ...

    @functools.wraps(func)
    def sync_wrapper(*args, **kwargs): ...

    if asyncio.iscoroutinefunction(func):
        return async_wrapper
    return sync_wrapper


class StructuredEvaluator(abc.ABC):
    def __new__(mcs, name, bases, namespace):
        if "evaluate" in namespace:
            namespace["evaluate"] = _trace_evaluation(namespace["evaluate"])

        return super().__new__(mcs, name, bases, namespace)

    @abc.abstractmethod
    def evaluate(
        self,
        *,
        system_prompt: Optional[str] = None,
        task_context: Optional[list[str]] = None,
        task_input: Optional[str] = None,
        task_output: Optional[str] = None,
        gold_answer: Optional[str] = None,
        task_metadata: Optional[dict[str, typing.Any]] = None,
        parent: Optional[typing.Any] = None,
        **kwargs,
    ) -> typing.Union[Optional[Evaluation], typing.Awaitable[Optional[Evaluation]]]: ...

    @abc.abstractmethod
    def explain(self, evaluation: Evaluation) -> typing.Union[Optional[str], typing.Awaitable[Optional[str]]]:
        return None


class RemoteEvaluator(StructuredEvaluator):
    def __init__(self, api, evaluator, criteria): ...

    def evaluate(
        self,
        *,
        system_prompt: Optional[str] = None,
        task_context: Optional[list[str]] = None,
        task_input: Optional[str] = None,
        task_output: Optional[str] = None,
        gold_answer: Optional[str] = None,
        task_metadata: Optional[dict[str, typing.Any]] = None,
        **kwargs,
    ) -> typing.Union[Optional[Evaluation], typing.Awaitable[Optional[Evaluation]]]: ...



import random

import time

import typing

from typing import Optional, Union, Any

import logging
import openai

from patronus import datasets, EvaluationResult
from patronus.evals import RemoteEvaluator, StructuredEvaluator
from patronus.experiments.experiment import run_experiment, Task
from patronus.experiments.types import EvalParent, TaskResult

oai = openai.OpenAI()

logging.getLogger("patronus").addHandler(logging.StreamHandler())

dataset = [
    {"sid": "f14", "task_input": "Say hello :)", "tags": """{"from_dataset": "true"}"""}
]

# EVALUATOR = "dummy-local-1"
# CRITERIA = "patronus:fifty-fifty"
# is_polite_evaluator = RemoteEvaluator(EVALUATOR, CRITERIA)
is_polite_evaluator = RemoteEvaluator("judge", "patronus:is-polite")


class DummyEvaluator(StructuredEvaluator):

    def evaluate(
        self,
        *,
        system_prompt: Optional[str] = None,
        task_context: Union[list[str], str, None] = None,
        task_attachments: Union[list[Any], None] = None,
        task_input: Optional[str] = None,
        task_output: Optional[str] = None,
        gold_answer: Optional[str] = None,
        task_metadata: Optional[typing.Dict[str, typing.Any]] = None,
        **kwargs: Any
    ) -> EvaluationResult:
        time.sleep(1)
        r = random.random()
        passed = r > 0.5
        return EvaluationResult(pass_=passed, score=r)


def answer(row: datasets.Row, **kwargs) -> TaskResult:
    return TaskResult(
        output="Hello! :) How can I assist you today?",
        tags={"task": "answer"},
    )
    # resp = oai.chat.completions.create(
    #     model="gpt-4o-mini",
    #     messages=[
    #     {
    #         "role": "user",
    #         "content": row.task_input,
    #     }
    # ])
    # return resp.choices[0].message.content


def answer_impolite(row, parent: EvalParent, tags) -> str:
    resp = oai.chat.completions.create(
        model="gpt-4o-mini",
        messages=[
            {"role": "system", "content": "Answer in impolite manner."},
            {
                "role": "user",
                "content": parent.task.output,
            },
        ],
    )
    return TaskResult(
        output=resp.choices[0].message.content,
        tags={
            "task": "answer_impolite",
            "model": "gpt-4o-mini",
        },
    )


ex = run_experiment(
    dataset=dataset,
    chain=[
        {
            "task": answer,
            "evaluators": [is_polite_evaluator, DummyEvaluator()],
        },
        {
            "task": answer_impolite,
            "evaluators": [is_polite_evaluator, DummyEvaluator()],
        },
    ],
    tags={
        "foo": "bar",
    },
)

ex.to_csv("./out.csv")


import datetime
import pydantic
from typing import Any, Optional
import yaml


class EvaluationResult(pydantic.BaseModel):
    """
    Container for evaluation outcomes including score, pass/fail status, explanations, and metadata.

    This class stores complete evaluation results with numeric scores, boolean pass/fail statuses,
    textual outputs, explanations, and arbitrary metadata. Evaluator functions can return instances
    of this class directly or return simpler types (bool, float, str) which will be automatically
    converted to EvaluationResult objects during recording.

    Attributes:
        score: Score of the evaluation. Can be any numerical value, though typically
            ranges from 0 to 1, where 1 represents the best possible score.
        pass_: Whether the evaluation is considered to pass or fail.
        text_output: Text output of the evaluation. Usually used for discrete
            human-readable category evaluation or as a label for score value.
        metadata: Arbitrary json-serializable metadata about evaluation.
        explanation: Human-readable explanation of the evaluation.
        tags: Key-value pair metadata.
        dataset_id: ID of the dataset associated with evaluated sample.
        dataset_sample_id: ID of the sample in a dataset associated with evaluated sample.
        evaluation_duration: Duration of the evaluation. In case value is not set,
            [@evaluator][patronus.evals.evaluators.evaluator] decorator and
            [Evaluator][patronus.evals.evaluators.Evaluator] classes will set this value automatically.
        explanation_duration: Duration of the evaluation explanation.
    """

    score: Optional[float] = None
    pass_: Optional[bool] = None
    text_output: Optional[str] = None
    metadata: Optional[dict[str, Any]] = None
    explanation: Optional[str] = None
    tags: Optional[dict[str, str]] = None
    dataset_id: Optional[str] = None
    dataset_sample_id: Optional[str] = None
    evaluation_duration: Optional[datetime.timedelta] = None
    explanation_duration: Optional[datetime.timedelta] = None

    def format(self) -> str:
        """
        Format the evaluation result into a readable summary.
        """
        md = self.model_dump(exclude_none=True, mode="json")
        return yaml.dump(md)

    def pretty_print(self, file=None) -> None:
        """
        Pretty prints the formatted content to the specified file or standard output.
        """
        f = self.format()
        print(f, file=file)



from patronus import evaluator
from patronus.experiments import FuncEvaluatorAdapter, run_experiment
from patronus.datasets import Row


@evaluator()
def exact_match(row: Row, **kwargs):
    return row.task_output == row.gold_answer


adapter = FuncEvaluatorAdapter(exact_match)

run_experiment(
    dataset=[{"task_output": "string", "gold_answer": "string"}], evaluators=[adapter]
)



from patronus import Client, evaluator, simple_task

cli = Client()


@evaluator
def iexact_match(evaluated_model_output: str, evaluated_model_gold_answer: str) -> bool:
    return evaluated_model_output.lower().strip() == evaluated_model_gold_answer.lower().strip()


eval_patronus_is_similar = cli.remote_evaluator(
    evaluator_id_or_alias="custom-small",
    criteria="system:is-similar-to-gold-answer",
)

cli.experiment(
    "mixed",
    data=[
        {"evaluated_model_input": "Foo", "evaluated_model_gold_answer": "hi foo"},
        {
            "evaluated_model_input": "Bar",
            "evaluated_model_gold_answer": "Hello bar",
        },
        {
            "evaluated_model_input": "Bar",
            "evaluated_model_gold_answer": "eloh Bar!",
        },
    ],
    task=simple_task(lambda input: f"Hi {input}"),
    evaluators=[iexact_match, eval_patronus_is_similar],
)



import patronus


cli = patronus.Client()

dataset = [
    {
        "evaluated_model_input": "1",
    },
    {
        "evaluated_model_input": "2",
    },
]


@patronus.task
def identity_task(row):
    return row.evaluated_model_input


@patronus.evaluator
def is_odd(task_result: patronus.TaskResult):
    return int(task_result.evaluated_model_output) % 2 == 1


@patronus.task
def stringify(parent: patronus.EvalParent):
    if parent.evals[is_odd].pass_:
        return "odd"
    else:
        return parent.task.evaluated_model_output


@patronus.evaluator
def is_number(task_result: patronus.TaskResult):
    try:
        int(task_result.evaluated_model_output)
        return True
    except ValueError:
        return False


ex = cli.experiment(
    "protego-dev",
    dataset=dataset,
    chain=[
        {"task": identity_task, "evaluators": [is_odd]},
        {"task": stringify, "evaluators": [is_number]},
    ],
)
df = ex.to_dataframe()
print(df)

df.to_csv("ex11.df.csv")



import logging
import random
import time

import patronus

#
# formatter = logging.Formatter("[%(levelname)-5s] [%(name)-10s] %(message)s")
# console_handler = logging.StreamHandler()
# console_handler.setLevel(logging.DEBUG)
# console_handler.setFormatter(formatter)
#
# plog = logging.getLogger("patronus")
# plog.setLevel(logging.DEBUG)
# plog.propagate = False
# plog.addHandler(console_handler)


cli = patronus.Client()

dataset = cli.remote_dataset("pii-questions-1.0.0")


@patronus.task
def local_task(row):
    time.sleep(random.random())
    return row.evaluated_model_input


@patronus.evaluator
def local_eval(row):
    time.sleep(random.random())
    return patronus.EvaluationResult(pass_=True, score_raw=random.random())


cli.experiment("protego-dev", dataset=dataset, task=local_task, evaluators=[local_eval])



from .clients import load_prompt as load_prompt
from .clients import aload_prompt as aload_prompt
from .clients import push_prompt as push_prompt
from .clients import apush_prompt as apush_prompt

# Exception classes
from .clients import PromptNotFoundError as PromptNotFoundError
from .clients import PromptProviderError as PromptProviderError
from .clients import PromptProviderConnectionError as PromptProviderConnectionError
from .clients import PromptProviderAuthenticationError as PromptProviderAuthenticationError

# Client classes
from .clients import PromptClient as PromptClient
from .clients import AsyncPromptClient as AsyncPromptClient

# Template engine classes
from .templating import TemplateEngine as TemplateEngine
from .templating import FStringTemplateEngine as FStringTemplateEngine
from .templating import MustacheTemplateEngine as MustacheTemplateEngine
from .templating import Jinja2TemplateEngine as Jinja2TemplateEngine

# Model classes
from .models import Prompt as Prompt
from .models import LoadedPrompt as LoadedPrompt



import inspect

import asyncio
import dataclasses
import json
import pathlib
import re
import typing
from typing import Optional, Union

import pandas as pd
import typing_extensions as te


class Attachment(typing.TypedDict, total=False):
    """
    Represent an attachment entry. Usually used in context of multimodal evaluation.
    """

    media_type: str
    url: str


class Fields(typing.TypedDict, total=False):
    """
    A TypedDict class representing fields for a structured data entity.

    Attributes:
        sid: An optional identifier for the system or session.
        system_prompt: An optional string representing the system
            prompt associated with the task.
        task_context: Optional contextual
            information for the task in the form of a string or a list of strings.
        task_attachments: Optional list of attachments associated with the task.
        task_input: An optional string representing the input data for the task. Usually a user input sent to an LLM.
        task_output: An optional string representing the output result of the task. Usually a response from an LLM.
        gold_answer: An optional string representing the correct or expected answer for evaluation purposes.
        task_metadata: Optional dictionary containing metadata associated with the task.
        tags: Optional dictionary holding additional key-value pair tags relevant to the task.
    """

    sid: te.NotRequired[Optional[str]]
    system_prompt: te.NotRequired[Optional[str]]
    task_context: te.NotRequired[Union[str, list[str], None]]
    task_attachments: te.NotRequired[Optional[list[Attachment]]]
    task_input: te.NotRequired[Optional[str]]
    task_output: te.NotRequired[Optional[str]]
    gold_answer: te.NotRequired[Optional[str]]
    task_metadata: te.NotRequired[Optional[dict[str, typing.Any]]]
    tags: te.NotRequired[Optional[dict[str, str]]]


@dataclasses.dataclass
class Row:
    """
    Represents a data row encapsulating access to properties in a pandas Series.

    Provides attribute-based access to underlying pandas Series data with properties
    that ensure compatibility with structured evaluators through consistent field naming
    and type handling.
    """

    _row: pd.Series

    def __getattr__(self, name: str):
        return self._row[name]

    @property
    def row(self) -> pd.Series:
        return self._row

    @property
    def dataset_id(self) -> Optional[str]:
        return self._row.get("dataset_id")

    @property
    def sid(self) -> str:
        return self._row.sid

    @property
    def system_prompt(self) -> Optional[str]:
        if "system_prompt" in self._row.index:
            return self._row.system_prompt
        return None

    @property
    def task_context(self) -> Optional[list[str]]:
        ctx = None
        if "task_context" in self._row.index:
            ctx = self._row.task_context
        if ctx is None:
            return None
        if isinstance(ctx, str):
            return [ctx]
        assert isinstance(ctx, list), f"task_context is not a list, its: {type(ctx)}"
        return ctx

    @property
    def task_attachments(self) -> Optional[list[typing.Any]]:
        attachments = None
        if "task_attachments" in self._row.index:
            attachments = self._row.task_attachments
        if attachments is None:
            return None
        return attachments

    @property
    def task_input(self) -> Optional[str]:
        if "task_input" in self._row.index:
            return self._row.task_input
        return None

    @property
    def task_output(self) -> Optional[str]:
        if "task_output" in self._row.index:
            return self._row.task_output
        return None

    @property
    def gold_answer(self) -> Optional[str]:
        if "gold_answer" in self._row.index:
            return self._row.gold_answer
        return None

    @property
    def task_metadata(self) -> Optional[dict[str, typing.Any]]:
        if "task_metadata" in self._row.index:
            return self._row.task_metadata
        return None

    @property
    def tags(self) -> Optional[dict[str, str]]:
        if "tags" in self._row.index:
            return self._row.tags
        return None


@dataclasses.dataclass
class Dataset:
    """
    Represents a dataset.
    """

    dataset_id: Optional[str]
    df: pd.DataFrame

    def iterrows(self) -> typing.Iterable[Row]:
        for i, row in self.df.iterrows():
            yield Row(row)

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame, dataset_id: Optional[str] = None) -> te.Self:
        df = cls.__sanitize_df(df, dataset_id)
        return cls(df=df, dataset_id=dataset_id)

    @classmethod
    def from_records(
        cls,
        records: Union[typing.Iterable[Fields], typing.Iterable[dict[str, typing.Any]]],
        dataset_id: Optional[str] = None,
    ) -> te.Self:
        """
        Creates an instance of the class by processing and sanitizing provided records
        and optionally associating them with a specific dataset ID.

        Args:
            records:
                A collection of records to initialize the instance. Each record can either
                be an instance of `Fields` or a dictionary containing corresponding data.
            dataset_id:
                An optional identifier for associating the data with a specific dataset.

        Returns:
            te.Self: A new instance of the class with the processed and sanitized data.
        """
        df = pd.DataFrame.from_records(records)
        df = cls.__sanitize_df(df, dataset_id)
        return cls(df=df, dataset_id=dataset_id)

    def to_csv(
        self, path_or_buf: Union[str, pathlib.Path, typing.IO[typing.AnyStr]], **kwargs: typing.Any
    ) -> Optional[str]:
        """
        Saves dataset to a CSV file.

        Args:
            path_or_buf: String path or file-like object where the CSV will be saved.
            **kwargs: Additional arguments passed to pandas.DataFrame.to_csv().

        Returns:
            String path if a path was specified and return_path is True, otherwise None.
        """
        return self.df.to_csv(path_or_buf, **kwargs)

    @classmethod
    def __sanitize_df(cls, df: pd.DataFrame, dataset_id: str) -> pd.DataFrame:
        # Validate and backfill "sid"
        if "sid" not in df.columns:
            df["sid"] = range(1, len(df) + 1)

        sid_count = df["sid"].count()
        if sid_count == 0:
            df["sid"] = range(1, len(df) + 1)

        if not pd.api.types.is_string_dtype(df["sid"]):
            try:
                df["sid"] = df["sid"].astype(str)
            except ValueError:
                raise ValueError("'sid' column contains non-integer values that cannot be converted to integers.")

        def normalize_context(value) -> Optional[list[str]]:
            if value is None:
                return None

            if isinstance(value, list):
                return [str(v) for v in value if v]

            if pd.isna(value) or value == "" or value == "nan":
                return None

            if isinstance(value, str):
                try:
                    parsed = json.loads(value)
                    if isinstance(parsed, list):
                        return [str(v) for v in parsed if v]
                    else:
                        return [str(parsed)]
                except json.JSONDecodeError:
                    return [value]

            return [str(value)]

        def _assert_attachment(value: dict):
            assert isinstance(
                value.get("url"), str
            ), "parsing 'task_attachments': missing or invalid type (expected str) of 'url' field"
            assert isinstance(
                value.get("media_type"), str
            ), "parsing 'task_attachments': missing or invalid type (expected str) of 'media_type' field"
            usage_type = value.get("usage_type")
            assert (
                isinstance(usage_type, str) or usage_type is None
            ), "parsing 'task_attachments': invalid type (expected str) of 'usage_type' field"

        def _normalize_attachment(value) -> dict[str, typing.Any]:
            if isinstance(value, dict):
                _assert_attachment(value)
                return value

        def normalize_attachments(value) -> Optional[list[dict[str, typing.Any]]]:
            if value is None:
                return None

            if isinstance(value, list):
                return [_normalize_attachment(v) for v in value]

            if isinstance(value, dict):
                _assert_attachment(value)
                return [value]

            if isinstance(value, str):
                try:
                    return normalize_attachments(json.loads(value))
                except json.JSONDecodeError as exc:
                    raise ValueError(f"parsing 'task_attachments': {exc}")

            raise ValueError("parsing 'task_attachments': unexpected value type")

        def normalize_metadata(value) -> Optional[dict[str, typing.Any]]:
            if value is None:
                return None
            if isinstance(value, dict):
                return value
            raise ValueError("parsing 'task_metadata': unexpected value type: expected dict or None")

        def normalize_tags(value) -> Optional[dict[str, str]]:
            if value is None:
                return None
            if isinstance(value, str):
                try:
                    return normalize_tags(json.loads(value))
                except json.JSONDecodeError as exc:
                    raise ValueError("parsing 'tags': unexpected value type") from exc
            if isinstance(value, dict):
                return value
            raise ValueError("parsing 'task_tags': unexpected value type: expected dict or None")

        if "system_prompt" in df.columns:
            df["system_prompt"] = df["system_prompt"].astype("string[python]")
            df["system_prompt"] = df["system_prompt"].replace({pd.NA: None})
        if "task_context" in df.columns:
            df["task_context"] = df["task_context"].apply(normalize_context)
            df["task_context"] = df["task_context"].astype("object")
        if "task_attachments" in df.columns:
            df["task_attachments"] = df["task_attachments"].apply(normalize_attachments)
            df["task_attachments"] = df["task_attachments"].astype("object")
        if "task_input" in df.columns:
            df["task_input"] = df["task_input"].astype("string[python]")
            df["task_input"] = df["task_input"].replace({pd.NA: None})
        if "task_output" in df.columns:
            df["task_output"] = df["task_output"].astype("string[python]")
            df["task_output"] = df["task_output"].replace({pd.NA: None})
        if "gold_answer" in df.columns:
            df["gold_answer"] = df["gold_answer"].astype("string[python]")
            df["gold_answer"] = df["gold_answer"].replace({pd.NA: None})
        if "task_metadata" in df.columns:
            df["task_metadata"] = df["task_metadata"].apply(normalize_metadata)
            df["task_metadata"] = df["task_metadata"].astype("object")
        if "tags" in df.columns:
            df["tags"] = df["tags"].apply(normalize_tags)
            df["tags"] = df["tags"].astype("object")

        # Backfill "dataset_id"
        if dataset_id:
            if "dataset_id" not in df.columns:
                df["dataset_id"] = dataset_id
            else:
                df["dataset_id"] = df["dataset_id"].fillna(dataset_id)

        df = df.sort_values("sid")
        return df


def read_csv(
    filename_or_buffer: Union[str, pathlib.Path, typing.IO[typing.AnyStr]],
    *,
    dataset_id: Optional[str] = None,
    sid_field: str = "sid",
    system_prompt_field: str = "system_prompt",
    task_input_field: str = "task_input",
    task_context_field: str = "task_context",
    task_attachments_field: str = "task_attachments",
    task_output_field: str = "task_output",
    gold_answer_field: str = "gold_answer",
    task_metadata_field: str = "task_metadata",
    tags_field: str = "tags",
    **kwargs: typing.Any,
) -> Dataset:
    """
    Reads a CSV file and converts it into a Dataset object. The CSV file is transformed
    into a structured dataset where each field maps to a specific aspect of the dataset
    schema provided via function arguments. You may specify custom field mappings as per
    your dataset structure, while additional keyword arguments are passed directly to the
    underlying 'pd.read_csv' function.

    Args:
        filename_or_buffer: Path to the CSV file or a file-like object containing the
            dataset to be read.
        dataset_id: Optional identifier for the dataset being read. Default is None.
        sid_field: Name of the column containing unique sample identifiers.
        system_prompt_field: Name of the column representing the system prompts.
        task_input_field: Name of the column containing the main input for the task.
        task_context_field: Name of the column describing the broader task context.
        task_attachments_field: Name of the column with supplementary attachments
            related to the task.
        task_output_field: Name of the column containing responses or outputs for the
            task.
        gold_answer_field: Name of the column detailing the expected or correct
            answer to the task.
        task_metadata_field: Name of the column storing metadata attributes
            associated with the task.
        tags_field: Name of the column containing tags or annotations related to each
            sample.
        **kwargs: Additional keyword arguments passed to 'pandas.read_csv' for fine-tuning
            the CSV parsing behavior, such as delimiters, encoding, etc.

    Returns:
        Dataset: The parsed dataset object containing structured data from the input
            CSV file.
    """
    return _read_dataframe(
        pd.read_csv,
        filename_or_buffer,
        dataset_id=dataset_id,
        sid_field=sid_field,
        system_prompt_field=system_prompt_field,
        task_context_field=task_context_field,
        task_attachments_field=task_attachments_field,
        task_input_field=task_input_field,
        task_output_field=task_output_field,
        gold_answer_field=gold_answer_field,
        task_metadata_field=task_metadata_field,
        tags_field=tags_field,
        **kwargs,
    )


def read_jsonl(
    filename_or_buffer: Union[str, pathlib.Path, typing.IO[typing.AnyStr]],
    *,
    dataset_id: Optional[str] = None,
    sid_field: str = "sid",
    system_prompt_field: str = "system_prompt",
    task_input_field: str = "task_input",
    task_context_field: str = "task_context",
    task_attachments_field: str = "task_attachments",
    task_output_field: str = "task_output",
    gold_answer_field: str = "gold_answer",
    task_metadata_field: str = "task_metadata",
    tags_field: str = "tags",
    **kwargs: typing.Any,
) -> Dataset:
    """
    Reads a JSONL (JSON Lines) file and transforms it into a Dataset object. This function
    parses the input data file or buffer in JSON Lines format into a structured format,
    extracting specified fields and additional metadata for usage in downstream tasks. The
    field mappings and additional keyword arguments can be customized to accommodate
    application-specific requirements.

    Args:
        filename_or_buffer: The path to the file or a file-like object containing the JSONL
            data to be read.
        dataset_id: An optional identifier for the dataset being read. Defaults to None.
        sid_field: The field name in the JSON lines representing the unique identifier for
            a sample. Defaults to "sid".
        system_prompt_field: The field name for the system prompt in the JSON lines file.
            Defaults to "system_prompt".
        task_input_field: The field name for the task input data in the JSON lines file.
            Defaults to "task_input".
        task_context_field: The field name for the task context data in the JSON lines file.
            Defaults to "task_context".
        task_attachments_field: The field name for any task attachments in the JSON lines
            file. Defaults to "task_attachments".
        task_output_field: The field name for task output data in the JSON lines file.
            Defaults to "task_output".
        gold_answer_field: The field name for the gold (ground truth) answer in the JSON
            lines file. Defaults to "gold_answer".
        task_metadata_field: The field name for metadata associated with the task in the
            JSON lines file. Defaults to "task_metadata".
        tags_field: The field name for tags in the parsed JSON lines file. Defaults to
            "tags".
        **kwargs: Additional keyword arguments to be passed to `pd.read_json` for
            customization. The parameter "lines" will be forcibly set to True if not
            provided.

    Returns:
        Dataset: A Dataset object containing the parsed and structured data.

    """
    kwargs.setdefault("lines", True)
    return _read_dataframe(
        pd.read_json,
        filename_or_buffer,
        dataset_id=dataset_id,
        sid_field=sid_field,
        system_prompt_field=system_prompt_field,
        task_context_field=task_context_field,
        task_attachments_field=task_attachments_field,
        task_input_field=task_input_field,
        task_output_field=task_output_field,
        gold_answer_field=gold_answer_field,
        task_metadata_field=task_metadata_field,
        tags_field=tags_field,
        **kwargs,
    )


def _read_dataframe(
    reader_function,
    filename_or_buffer: Union[str, pathlib.Path, typing.IO[typing.AnyStr]],
    *,
    dataset_id: Optional[str] = None,
    sid_field: str = "sid",
    system_prompt_field: str = "system_prompt",
    task_context_field: str = "task_context",
    task_attachments_field: str = "task_attachments",
    task_input_field: str = "task_input",
    task_output_field: str = "task_output",
    gold_answer_field: str = "gold_answer",
    task_metadata_field: str = "task_metadata",
    tags_field: str = "tags",
    **kwargs: typing.Any,
) -> Dataset:
    df = reader_function(filename_or_buffer, **kwargs)

    if sid_field in df.columns:
        df["sid"] = df[sid_field]
    if system_prompt_field in df.columns:
        df["system_prompt"] = df[system_prompt_field]
    if task_context_field in df.columns:
        df["task_context"] = df[task_context_field]
    if task_attachments_field in df.columns:
        df["task_attachments"] = df[task_metadata_field]
    if task_input_field in df.columns:
        df["task_input"] = df[task_input_field]
    if task_output_field in df.columns:
        df["task_output"] = df[task_output_field]
    if gold_answer_field in df.columns:
        df["gold_answer"] = df[gold_answer_field]
    if task_metadata_field in df.columns:
        df["task_metadata"] = df[task_metadata_field]
    if tags_field in df.columns:
        df["tags"] = df[tags_field]

    dataset_id = _sanitize_dataset_id(dataset_id)
    return Dataset.from_dataframe(df, dataset_id=dataset_id)


def _sanitize_dataset_id(dataset_id: str) -> Optional[str]:
    if not dataset_id:
        return None
    dataset_id = re.sub(r"[^a-zA-Z0-9\-_]", "-", dataset_id.strip())
    if not dataset_id:
        return None
    return dataset_id


class DatasetLoader:
    """
    Encapsulates asynchronous loading of a dataset.

    This class provides a mechanism to lazily load a dataset asynchronously only
    once, using a provided dataset loader function.
    """

    def __init__(self, loader: Union[typing.Awaitable[Dataset], typing.Callable[[], typing.Awaitable[Dataset]]]):
        self.__lock = asyncio.Lock()
        self.__loader = loader
        self.dataset: Optional[Dataset] = None

    async def load(self) -> Dataset:
        """
        Load dataset. Repeated calls will return already loaded dataset.
        """
        async with self.__lock:
            if self.dataset is not None:
                return self.dataset
            if inspect.iscoroutinefunction(self.__loader):
                self.dataset = await self.__loader()
            else:
                self.dataset = await self.__loader
            return self.dataset



import patronus


patronus.init(
    # otel_endpoint="https://otel.patronus.ai:4318",
    otel_endpoint="http://0.0.0.0:4318",
    otel_exporter_otlp_protocol="http/protobuf",
)


@patronus.traced()
def main(a):
    return f"{a} world"


main("hello")



import asyncio
import typing

T = typing.TypeVar("T")


def run_until_complete(coro: typing.Awaitable[T]) -> typing.Awaitable[T]:
    try:
        loop = asyncio.get_running_loop()
        return loop.create_task(coro)
    except RuntimeError:
        # Ignore this exception since we can continue by creating an event loop.
        # We are not creating an event loop in the except block to no include
        # this stacktrace part in case of a raised exception in the coroutine.
        # Otherwise, we would get message as
        #
        # > RuntimeError: no running event loop
        # > During handling of the above exception, another exception occurred: [...]
        #
        # Which we don't want because the try catch is just a check whether there is a running event loop.
        # We are not really handling an "exception" here.
        pass

    loop = asyncio.get_event_loop()
    return loop.run_until_complete(coro)



from patronus.experiments import run_experiment
from patronus.evals import RemoteEvaluator

run_experiment(
    project_name="Tutorial",
    dataset=[
        {
            "task_input": "Please provide your contact details.",
            "task_output": "My email is john.doe@example.com and my phone number is 123-456-7890.",
        },
        {
            "task_input": "Share your personal information.",
            "task_output": "My name is Jane Doe and I live at 123 Elm Street.",
        },
    ],
    evaluators=[RemoteEvaluator("pii")],
    experiment_name="Detect PII",
)



#!/usr/bin/env -S uv run --script
# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "patronus",
#     "pydantic-ai-slim[openai]",
#     "opentelemetry-instrumentation-asyncio>=0.54b0,<1.0",
#     "opentelemetry-instrumentation-threading>=0.54b0,<1.0",
# ]
# ///
import asyncio
from pydantic_ai import Agent
from opentelemetry.instrumentation.threading import ThreadingInstrumentor
from opentelemetry.instrumentation.asyncio import AsyncioInstrumentor
from patronus.integrations.pydantic_ai import PydanticAIIntegrator
import patronus

patronus.init(
    integrations=[
        PydanticAIIntegrator(),
        ThreadingInstrumentor(),
        AsyncioInstrumentor(),
    ]
)


def get_agent(system_prompt="You are a helpful assistant"):
    agent = Agent("openai:gpt-4o", output_type=str, system_prompt=system_prompt)
    return agent


@patronus.traced("weather-pydantic-ai")
async def main():
    # Create weather agent and attach tool to it
    weather_agent = get_agent(
        "You are a helpful assistant that can help with weather information."
    )

    @weather_agent.tool_plain()
    async def get_weather():
        # Mock tool output
        return (
            "Today's weather is Sunny with a forecasted high of 30°C and a low of 25°C. "
            "The wind is expected at 4 km/h."
        )

    # Create manager agent
    manager_agent = get_agent(
        "You are a helpful assistant that can coordinate with other subagents "
        "and query them for more information about topics."
    )

    # Create a tool to execute the weather agent
    @manager_agent.tool_plain()
    async def call_weather_agent():
        weather_info = await weather_agent.run("What is the weather in Paris, France?")
        return str(weather_info)

    # Run the manager
    print("Running the agent...")
    return await manager_agent.run("What is the weather in Paris, France?")


if __name__ == "__main__":
    result = asyncio.run(main())
    print(result)



from patronus import Client

client = Client()

evaluator = client.remote_evaluator(EVALUATOR, PROFILE_NAME, max_attempts=3)

await evaluator.evaluate(
    evaluated_model_system_prompt="SYSTEM_PROMPT",
    evaluated_model_retrieved_context=["RETRIEVED_CONTEXT"],
    evaluated_model_input="MODEL_INPUT",
    evaluated_model_output="MODEL_OUTPUT",
    evaluated_model_gold_answer="GOLD_ANSWER",
    app="playground",
)



import statistics

import pandas as pd
import typing
import asyncio
import sys
import traceback
from typing import Optional

from .types import TaskResult
from .tqdm import AsyncTQDMWithHandle

from patronus import datasets
from patronus.exceptions import WithStackTraceMixin
from patronus.evals import EvaluationResult


class Result(typing.NamedTuple):
    link_idx: int
    task_name: Optional[str]
    task_result: Optional[TaskResult]
    evaluator_id: str
    criteria: Optional[str]
    evaluation_result: Optional[EvaluationResult]
    row: datasets.Row


class Reporter:
    tqdm: Optional[AsyncTQDMWithHandle]

    def __init__(self):
        self.tqdm = None
        self.lock = asyncio.Lock()

        self.results: list[Result] = []
        self.df: Optional[pd.DataFrame] = None

    def set_tqdm(self, tqdm: AsyncTQDMWithHandle):
        self.tqdm = tqdm

    def add_task_error(self, exc: Exception, row: datasets.Row):
        stack_trace = self._get_stack_trace(exc)
        self.print_error(
            "\n".join(
                [stack_trace, f"Task failed on sample {(row.dataset_id, row.sid)=} with the following error: {exc}"]
            )
        )

    def add_evaluator_error(self, exc: Exception, row: datasets.Row, evaluator_id: str, criteria: Optional[str]):
        stack_trace = self._get_stack_trace(exc)
        self.print_error(
            "\n".join(
                [
                    stack_trace,
                    f"Evaluator {(evaluator_id, criteria)=} failed on sample "
                    f"{(row.dataset_id, row.sid)=} with the following error: {exc}",
                ]
            )
        )

    @staticmethod
    def _get_stack_trace(exc: Exception) -> Optional[str]:
        stack_trace = None
        if isinstance(exc, WithStackTraceMixin):
            stack_trace = exc.stack_trace
        if stack_trace is None:
            tb_lines = traceback.format_exception(type(exc), exc, exc.__traceback__)
            tb_text = "".join(tb_lines)
            stack_trace = tb_text
        return stack_trace

    def print_error(self, message: str):
        if self.tqdm:
            self.tqdm.clear()
        print(message, file=sys.stderr)
        if self.tqdm:
            self.tqdm.display()

    async def add_result(
        self,
        link_idx: int,
        task_name: Optional[str],
        task_result: Optional[TaskResult],
        evaluator_id: str,
        criteria: Optional[str],
        evaluation_result: Optional[EvaluationResult],
        row: datasets.Row,
    ):
        async with self.lock:
            self.results.append(
                Result(
                    link_idx=link_idx,
                    task_name=task_name,
                    task_result=task_result,
                    evaluator_id=evaluator_id,
                    criteria=criteria,
                    evaluation_result=evaluation_result,
                    row=row,
                )
            )

    def to_dataframe(self) -> pd.DataFrame:
        if self.df is not None:
            return self.df

        data = []

        for result in self.results:
            row_data = {
                "link_idx": result.link_idx,
                "task.name": result.task_name,
                "evaluator_id": result.evaluator_id,
                "criteria": result.criteria,
            }

            if result.task_result:
                row_data.update(
                    {
                        "task.output": result.task_result.output,
                        "task.metadata": result.task_result.metadata,
                        "task.tags": result.task_result.tags,
                    }
                )
            else:
                row_data.update(
                    {
                        "task.output": None,
                        "task.metadata": None,
                        "task.tags": None,
                    }
                )

            if result.evaluation_result:
                row_data.update(
                    {
                        "eval.score": result.evaluation_result.score,
                        "eval.pass": result.evaluation_result.pass_,
                        "eval.text_output": result.evaluation_result.text_output,
                        "eval.metadata": result.evaluation_result.metadata,
                        "eval.explanation": result.evaluation_result.explanation,
                        "eval.tags": result.evaluation_result.tags,
                        "eval.evaluation_duration": result.evaluation_result.evaluation_duration,
                        "eval.explanation_duration": result.evaluation_result.explanation_duration,
                    }
                )
            else:
                row_data.update(
                    {
                        "eval.score": None,
                        "eval.pass": None,
                        "eval.text_output": None,
                        "eval.metadata": None,
                        "eval.explanation": None,
                        "eval.tags": None,
                        "eval.evaluation_duration": None,
                        "eval.explanation_duration": None,
                    }
                )

            if result.row:
                row_series = result.row.row

                for column_name, value in row_series.items():
                    # Avoid overwriting existing columns
                    if column_name not in row_data:
                        row_data[column_name] = value

            data.append(row_data)

        self.df = pd.DataFrame(data)
        # reclaim memory
        self.results = []
        return self.df

    def summary(self):
        # summary can be called only once experiment finished.
        # We set tqdm to None to prevent showing progress bar in between error messages
        self.tqdm = None
        df = self.to_dataframe()

        if df.empty:
            self.print_error("\nEvaluation summary could not be generated: No successful evaluations found.")
            self.print_error(
                "All evaluations in the experiment may have failed. See the error messages above for details."
            )
            return

        df["criteria"] = df["criteria"].fillna("None")

        grouped = df.groupby(["link_idx", "evaluator_id", "criteria"])

        def group_name(evaluator_id, criteria, link_idx):
            if criteria != "None":
                ev = f"{criteria} ({evaluator_id})"
            else:
                ev = evaluator_id
            return f"{ev} [link_idx={int(link_idx)}]"

        for (link_idx, evaluator_id, criteria), group in grouped:
            name = group_name(evaluator_id, criteria, link_idx)

            scores_series = group["eval.score"]
            passes_series = group["eval.pass"]

            scores_and_passes = list(zip(scores_series, passes_series))

            scores = [s for s, _ in scores_and_passes if s is not None]
            passes = [int(p) for _, p in scores_and_passes if p is not None]

            if scores:
                print_summary(name, scores, passes, len(scores_and_passes), display_hist=True)


def print_summary(name: str, scores: list[float], passes: list[int], count: int, display_hist: bool):
    print()
    print(name)
    print("-" * len(name))
    print(f"Count     : {count}")
    print(f"Pass rate : {round(statistics.mean(passes), 3)}")
    print(f"Mean      : {round(statistics.mean(scores), 3)}")
    print(f"Min       : {round(min(scores), 3)}")
    print(f"25%       : {round(percentile(scores, 25), 3)}")
    print(f"50%       : {round(percentile(scores, 50), 3)}")
    print(f"75%       : {round(percentile(scores, 75), 3)}")
    print(f"Max       : {round(max(scores), 3)}")

    if display_hist:
        print()
        print("Score distribution")
        print_histogram(scores)


def percentile(data: list[float], p: int):
    data = sorted(data)
    index = (p / 100) * (len(data) - 1)
    if index.is_integer():
        return data[int(index)]
    else:
        lower_bound = int(index)
        upper_bound = lower_bound + 1
        weight = index - lower_bound
        return data[lower_bound] * (1 - weight) + data[upper_bound] * weight


def print_histogram(data, bin_count=5):
    # Calculate the range of the data
    min_val = min(data)
    max_val = max(data)

    if min_val == max_val:
        if min_val > 0.5:
            min_val = 0
        else:
            max_val = 1

    range_val = max_val - min_val

    # Calculate bin size
    bin_size = range_val / bin_count

    # Initialize bins
    bins = [0] * bin_count

    # Distribute data into bins
    for value in data:
        # Find the appropriate bin for the current value
        bin_index = int((value - min_val) / bin_size)
        # Edge case for the maximum value
        if bin_index == bin_count:
            bin_index -= 1
        bins[bin_index] += 1

    # Determine the width of the histogram
    max_bin_count = max(bins)
    scale_factor = 20 / max_bin_count  # Scale the histogram to a max width of 50 characters

    # Print the histogram
    print("Score Range".ljust(20), "Count".ljust(10), "Histogram")
    for i in range(bin_count):
        bin_start = min_val + i * bin_size
        bin_end = bin_start + bin_size
        bin_count = bins[i]
        bar = "#" * int(bin_count * scale_factor)
        print(f"{bin_start:.2f} - {bin_end:.2f}".ljust(20), f"{bin_count}".ljust(10), bar)



import logging

import asyncio
import datetime
import functools
import inspect
import random
import time
import traceback
from typing import Optional

from patronus.api.api_client_base import RetryError, RPMLimitError, UnrecoverableAPIError

log = logging.getLogger("patronus.core")


def retry(max_attempts=3, initial_delay=1, backoff_factor=2):
    def decorator(func):
        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def wrapper(*args, **kwargs):
                attempts = 1
                delay = initial_delay
                last_error = None
                stack_trace: Optional[str] = None

                while attempts <= max_attempts:
                    call_time = datetime.datetime.now()
                    try:
                        return await func(*args, **kwargs)
                    except UnrecoverableAPIError as e:
                        raise RetryError(attempts, max_attempts, e, traceback.format_exc())
                    except RPMLimitError as err:
                        last_error = err
                        stack_trace = traceback.format_exc()
                        log.debug(f"api_retry: Attempt {attempts} out of {max_attempts}: {err}")
                        wait_for_s = err.wait_for_s
                        if not wait_for_s:
                            now = datetime.datetime.now()
                            if call_time.minute != now.minute:
                                continue
                            next_minute = now.replace(second=0, microsecond=0) + datetime.timedelta(minutes=1)
                            wait_for_s = (next_minute - now).total_seconds() + (3 + random.random() * 5)
                        if attempts < max_attempts:
                            log.debug(f"Attempting again in {wait_for_s}s.")
                            await asyncio.sleep(wait_for_s)
                    except Exception as err:
                        log.debug(f"api_retry: Attempt {attempts} out of {max_attempts}: {err}")
                        last_error = err
                        stack_trace = traceback.format_exc()
                        if attempts < max_attempts:
                            await asyncio.sleep(delay)
                    finally:
                        delay *= backoff_factor
                        attempts += 1

                raise RetryError(attempts - 1, max_attempts, last_error, stack_trace) from last_error

            return wrapper
        else:

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                attempts = 1
                delay = initial_delay
                last_error = None
                stack_trace: Optional[str] = None

                while attempts <= max_attempts:
                    call_time = datetime.datetime.now()
                    try:
                        return func(*args, **kwargs)
                    except UnrecoverableAPIError as e:
                        raise RetryError(attempts, max_attempts, e, traceback.format_exc())
                    except RPMLimitError as err:
                        last_error = err
                        stack_trace = traceback.format_exc()
                        log.debug(f"api_retry: Attempt {attempts} out of {max_attempts}: {err}")
                        wait_for_s = err.wait_for_s
                        if not wait_for_s:
                            now = datetime.datetime.now()
                            if call_time.minute != now.minute:
                                continue
                            next_minute = now.replace(second=0, microsecond=0) + datetime.timedelta(minutes=1)
                            wait_for_s = (next_minute - now).total_seconds() + (3 + random.random() * 5)
                        if attempts < max_attempts:
                            log.debug(f"Attempting again in {wait_for_s}s.")
                            time.sleep(wait_for_s)
                    except Exception as err:
                        log.debug(f"api_retry: Attempt {attempts} out of {max_attempts}: {err}")
                        last_error = err
                        stack_trace = traceback.format_exc()
                        if attempts < max_attempts:
                            time.sleep(delay)
                    finally:
                        delay *= backoff_factor
                        attempts += 1

                raise RetryError(attempts - 1, max_attempts, last_error, stack_trace) from last_error

            return wrapper

    return decorator



import patronus
from patronus.datasets import Row


def test_non_parametrized_evaluator_decorator_use():
    @patronus.evaluator
    def ematch(row: Row):
        return row.task_input == row.task_output

    assert ematch.display_name() == "ematch"


def test_parametrized_evaluator_decorator_use():
    eval_name, profile_name = "my-evaluator", "profile1"

    @patronus.evaluator(name=eval_name, criteria=profile_name)
    def ematch(row: Row):
        return row.task_input == row.task_output

    assert ematch.display_name() == f"{eval_name}:{profile_name}"



#!/usr/bin/env python3
"""Test the reorganized architecture."""

import tempfile
import sys
import os

# Add the src directory to the Python path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'src'))

def test_architecture():
    """Test the new architecture without circular imports."""
    try:
        # Test storage imports
        from patronus.prompts.storage import LocalPromptStorage, StorageInfo
        print("✓ Storage imports successful")
        
        # Test provider imports
        from patronus.prompts.providers import LocalPromptProvider, APIPromptProvider
        print("✓ Provider imports successful")
        
        # Test CLI imports
        from patronus.cli.commands.prompts import pull, tidy
        print("✓ CLI command imports successful")
        
        # Test provider factory from clients
        from patronus.prompts.clients import PromptClient
        print("✓ Client imports successful")
        
        # Test storage instantiation
        with tempfile.TemporaryDirectory() as temp_dir:
            storage = LocalPromptStorage(temp_dir)
            print("✓ Storage instantiation successful")
            
            # Test provider instantiation
            provider = LocalPromptProvider(temp_dir)
            print("✓ Provider instantiation successful")
            
            # Test client with new providers
            client = PromptClient()
            print("✓ Client instantiation successful")
        
        print("\n✅ All tests passed! Architecture is correct.")
        return True
        
    except ImportError as e:
        print(f"❌ Import error: {e}")
        return False
    except Exception as e:
        print(f"❌ Error: {e}")
        return False

if __name__ == "__main__":
    success = test_architecture()
    sys.exit(0 if success else 1)



from patronus import Client, evaluator, Row

client = Client()


@evaluator
def iexact_match(row: Row) -> bool:
    return row.evaluated_model_output.lower().strip() == row.evaluated_model_gold_answer.lower().strip()


client.experiment(
    "Tutorial",
    dataset=[
        {
            "evaluated_model_input": "Translate 'Good night' to French.",
            "evaluated_model_output": "bonne nuit",
            "evaluated_model_gold_answer": "Bonne nuit",
        },
        {
            "evaluated_model_input": "Summarize: 'AI improves efficiency'.",
            "evaluated_model_output": "ai improves efficiency",
            "evaluated_model_gold_answer": "AI improves efficiency",
        },
    ],
    evaluators=[iexact_match],
    experiment_name="Case Insensitive Match",
)



#!/usr/bin/env -S uv run --script
# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "patronus",
#     "openai",
#     "openinference-instrumentation-openai>=0.1.28,<0.2.0",
#     "opentelemetry-instrumentation-threading>=0.54b0,<1.0",
#     "opentelemetry-instrumentation-asyncio>=0.54b0,<1.0"
# ]
# ///
import json

import requests
from openai import OpenAI
from openinference.instrumentation.openai import OpenAIInstrumentor
import patronus

# Initialize patronus with OpenAI Instrumentor
patronus.init(integrations=[OpenAIInstrumentor()])


def get_weather(latitude, longitude):
    response = requests.get(
        f"https://api.open-meteo.com/v1/forecast?latitude={latitude}&longitude={longitude}&current=temperature_2m,wind_speed_10m&hourly=temperature_2m,relative_humidity_2m,wind_speed_10m"
    )
    data = response.json()
    return data["current"]["temperature_2m"]


def get_client():
    client = OpenAI()
    return client


@patronus.traced()
def call_llm(client, user_prompt):
    tools = [
        {
            "type": "function",
            "name": "get_weather",
            "description": "Get current temperature for provided coordinates in celsius.",
            "parameters": {
                "type": "object",
                "properties": {
                    "latitude": {"type": "number"},
                    "longitude": {"type": "number"},
                },
                "required": ["latitude", "longitude"],
                "additionalProperties": False,
            },
            "strict": True,
        }
    ]

    input_messages = [{"role": "user", "content": user_prompt}]

    response = client.responses.create(
        model="gpt-4.1",
        input=input_messages,
        tools=tools,
    )
    return response


@patronus.traced("openai-weather")
def main():
    user_prompt = "What's the weather like in Paris today?"

    client = get_client()
    response = call_llm(client, user_prompt)
    print("LLM Response")
    print(response.model_dump_json())

    weather_response = None
    if response.output:
        output = response.output[0]
        if output.type == "function_call" and output.name == "get_weather":
            kwargs = json.loads(output.arguments)
            print("Weather API Response")
            weather_response = get_weather(**kwargs)
            print(weather_response)

    if weather_response:
        print(user_prompt)
        print(f"Answer: {weather_response}")


if __name__ == "__main__":
    main()



import asyncio
import inspect
import logging
import typing
from typing import Optional

import typing_extensions as te

from . import api, api_types, evaluators, types
from .datasets import Row
from .retry import retry

log = logging.getLogger(__name__)


class RemoteEvaluatorError(evaluators.EvaluatorError):
    def __init__(self, evaluator: str, criteria: str, status: str, error_message: str):
        super().__init__(
            f"{evaluator!r} with criteria {criteria!r} "
            f"returned unexpected status {status!r} with error message {error_message!r}"
        )


class EvaluateCall(typing.Protocol):
    def __call__(
        self,
        evaluated_model_system_prompt: Optional[str] = None,
        evaluated_model_retrieved_context: Optional[list[str]] = None,
        evaluated_model_input: Optional[str] = None,
        evaluated_model_output: Optional[str] = None,
        evaluated_model_gold_answer: Optional[str] = None,
        evaluated_model_attachments: Optional[list[dict[str, typing.Any]]] = None,
    ) -> typing.Awaitable[api_types.EvaluationResult]: ...


class EvaluateWrapperFunction(typing.Protocol):
    def __call__(
        self,
        call: EvaluateCall,
        *,
        row: Row,
        task_result: typing.Optional[types.TaskResult] = None,
        evaluated_model_system_prompt: Optional[str] = None,
        evaluated_model_retrieved_context: Optional[list[str]] = None,
        evaluated_model_input: Optional[str] = None,
        evaluated_model_output: Optional[str] = None,
        evaluated_model_gold_answer: Optional[str] = None,
        evaluated_model_attachments: Optional[list[dict[str, typing.Any]]] = None,
        parent: types.EvalParent = None,
        **kwargs,
    ) -> typing.Awaitable[api_types.EvaluateResponse]: ...


class RemoteEvaluator(evaluators.Evaluator):
    remote_capture = True

    def __init__(
        self,
        evaluator_id_or_alias: str,
        criteria: str,
        *,
        explain_strategy: typing.Literal["never", "on-fail", "on-success", "always"] = "always",
        criteria_config: Optional[dict[str, typing.Any]] = None,
        allow_update: bool = False,
        # Maximum number of attempts in case when evaluation throws an exception.
        max_attempts: int = 3,
        api_: api.PatronusAPIClient,
        override_func: Optional[EvaluateWrapperFunction] = None,
    ):
        self.__lock = asyncio.Lock()
        self.__initialized = False

        self.name = evaluator_id_or_alias
        self.evaluator = evaluator_id_or_alias

        self.criteria = criteria

        self.explain_strategy = explain_strategy
        self.criteria_config = criteria_config
        self.allow_update = allow_update
        self.max_attempts = max_attempts
        self.api = api_

        self.override_func = override_func

        super().__init__(evaluators.EVALUATION_ARGS)

    def set_evaluator_ref(self, evaluator: str):
        self.name = evaluator
        self.evaluator = evaluator

    async def load(self) -> te.Self:
        async with self.__lock:
            if self.__initialized:
                return

            evs = await self.api.list_evaluators()

            ev: Optional[api_types.Evaluator] = None
            for e in evs:
                if e.id == self.evaluator:
                    ev = e
                for alias in e.aliases:
                    if alias == self.evaluator:
                        ev = e

            if ev is None:
                raise ValueError(f"Evaluator {self.evaluator!r} not found")

            if self.criteria_config:
                await self._init_from_config(ev)
            else:
                await self._init_existing(ev=ev)

            self.__initialized = True
        return self

    async def _init_from_config(self, ev: api_types.Evaluator):
        if not self.criteria:
            raise ValueError("criteria is required when specifying criteria_config")
        if self.criteria.startswith("system:"):
            raise ValueError(
                f"Cannot use criteria_config with patronus criteria. Provided criteria was {self.criteria!r}"
            )

        criteria_list = (
            await self.api.list_criteria(
                api_types.ListCriteriaRequest(
                    evaluator_family=ev.evaluator_family,
                    name=self.criteria,
                    get_last_revision=True,
                )
            )
        ).evaluator_criteria

        if not criteria_list:
            log.info(
                f"No evaluator criteria {self.criteria!r} for evaluator {ev.evaluator_family!r} found. Creating one..."
            )
            criteria = (
                await self.api.create_criteria(
                    api_types.CreateCriteriaRequest(
                        evaluator_family=ev.evaluator_family,
                        name=self.criteria,
                        config=self.criteria_config,
                    )
                )
            ).evaluator_criteria
            log.info(f"Evaluator criteria {self.criteria} created for evaluator family {ev.evaluator_family}.")
        elif len(criteria_list) > 1:
            raise Exception(
                f"Unexpected number of criteria retrieved for "
                f"evaluator {self.evaluator!r} and criteria name {self.criteria!r}"
            )
        else:
            criteria = criteria_list[0]

        # Check if user provided criteria config is subset of existing config
        # This checks only one level of the config, but we don't support criteria with nested
        # structure at this point so, it's alright.
        is_subset = {**criteria.config, **self.criteria_config} == criteria.config

        if not is_subset and not self.allow_update:
            raise ValueError(
                "Provided 'criteria_config' differs from existing criteria. "
                "Please set 'allow_update=True' if you wish to update the criteria. "
                "Updating criteria can be unsafe if they're used in production system or by other people."
            )

        if not is_subset:
            log.info("Existing criteria config differs from the provided config. Adding revision to the criteria...")
            criteria_resp = await self.api.add_evaluator_criteria_revision(
                criteria.public_id,
                api_types.AddEvaluatorCriteriaRevisionRequest(
                    config={**criteria.config, **self.criteria_config},
                ),
            )
            log.info(f"Revision added to evaluator criteria {criteria_resp.evaluator_criteria.name}.")

        self.set_evaluator_ref(ev.id)

    async def _init_existing(self, ev: api_types.Evaluator):
        criteria_list = await self.api.list_criteria(
            api_types.ListCriteriaRequest(
                evaluator_family=ev.evaluator_family,
                name=self.criteria,
                get_last_revision=True,
            )
        )
        if len(criteria_list.evaluator_criteria) == 0:
            raise ValueError(f"Criteria for evaluator {self.evaluator!r} given name {self.criteria!r} not found")
        if len(criteria_list.evaluator_criteria) > 1:
            raise ValueError(f"More than 1 criteria found for evaluator {self.evaluator!r}")

        self.criteria = criteria_list.evaluator_criteria[0].name
        self.set_evaluator_ref(ev.id)

    async def evaluate(
        self,
        *,
        evaluated_model_system_prompt: Optional[str] = None,
        evaluated_model_retrieved_context: Optional[list[str]] = None,
        evaluated_model_input: Optional[str] = None,
        evaluated_model_output: Optional[str] = None,
        evaluated_model_gold_answer: Optional[str] = None,
        evaluated_model_attachments: Optional[list[dict[str, typing.Any]]] = None,
        tags: Optional[dict[str, str]] = None,
        dataset_id: Optional[str] = None,
        dataset_sample_id: Optional[int] = None,
        app: Optional[str] = None,
        # experiment_id is generally used by the framework.
        experiment_id: Optional[str] = None,
        # row is framework specific field. It's not necessary when using evaluator outside of framework.
        row: Optional[Row] = None,
        # task_result is framework specific field. It's not necessary when using evaluator outside of framework.
        task_result: Optional[types.TaskResult] = None,
        # parent is framework specific field. It's not necessary when using evaluator outside of framework.
        parent: types.EvalParent = None,
        **kwargs,
    ) -> api_types.EvaluationResult:
        # Make sure that evaluator is loaded
        await self.load()

        @retry(max_attempts=self.max_attempts)
        async def call(
            evaluated_model_system_prompt: Optional[str] = None,
            evaluated_model_retrieved_context: Optional[list[str]] = None,
            evaluated_model_input: Optional[str] = None,
            evaluated_model_output: Optional[str] = None,
            evaluated_model_gold_answer: Optional[str] = None,
            evaluated_model_attachments: Optional[list[dict[str, typing.Any]]] = None,
        ) -> api_types.EvaluationResult:
            return await self.api.evaluate_one(
                api_types.EvaluateRequest(
                    evaluators=[
                        api_types.EvaluateEvaluator(
                            evaluator=self.evaluator,
                            criteria=self.criteria,
                            explain_strategy=self.explain_strategy,
                        )
                    ],
                    evaluated_model_system_prompt=evaluated_model_system_prompt,
                    evaluated_model_retrieved_context=evaluated_model_retrieved_context,
                    evaluated_model_input=evaluated_model_input,
                    evaluated_model_output=evaluated_model_output,
                    evaluated_model_gold_answer=evaluated_model_gold_answer,
                    evaluated_model_attachments=evaluated_model_attachments,
                    app=app,
                    experiment_id=experiment_id,
                    capture="all",
                    dataset_id=dataset_id,
                    dataset_sample_id=dataset_sample_id,
                    tags=tags,
                )
            )

        if self.override_func is not None:
            try:
                response = self.override_func(
                    call,
                    row=row,
                    task_result=task_result,
                    evaluated_model_system_prompt=evaluated_model_system_prompt,
                    evaluated_model_retrieved_context=evaluated_model_retrieved_context,
                    evaluated_model_input=evaluated_model_input,
                    evaluated_model_output=evaluated_model_output,
                    evaluated_model_gold_answer=evaluated_model_gold_answer,
                    evaluated_model_attachments=evaluated_model_attachments,
                    parent=parent,
                )
            except TypeError as exc:
                if "got an unexpected keyword argument" in "".join(exc.args):
                    raise TypeError(f"{exc}. Did you forgot to add '**kwargs' to the function parameters?")
                raise exc
            if inspect.iscoroutine(response):
                response = await response

            if response is None:
                return None
        else:
            response = await call(
                evaluated_model_system_prompt=evaluated_model_system_prompt,
                evaluated_model_retrieved_context=evaluated_model_retrieved_context,
                evaluated_model_input=evaluated_model_input,
                evaluated_model_output=evaluated_model_output,
                evaluated_model_gold_answer=evaluated_model_gold_answer,
                evaluated_model_attachments=evaluated_model_attachments,
            )

        return response

    def wrap(self, fn: EvaluateWrapperFunction) -> "RemoteEvaluator":
        return RemoteEvaluator(
            evaluator_id_or_alias=self.name,
            criteria=self.criteria,
            explain_strategy=self.explain_strategy,
            criteria_config=self.criteria_config,
            allow_update=self.allow_update,
            max_attempts=self.max_attempts,
            api_=self.api,
            override_func=fn,
        )
